;;;; Created on 2009-10-26 16:57:19

(in-package :weblocks-postmodern)

(export '(order-by-expression range-to-offset range-to-limit))

(defparameter *oid-p* nil)

;;Works like with-connection but uses an existing connection.
;;TODO: Should this just be replaced with postmodern:with-connection??
(defmacro with-store (store &body body)
  "Connects the connection if it is disconnected. Does body and then disconnects again."
  `(let ((postmodern:*database* (if (cl-postgres:database-open-p ,store)
				    ,store
				    (progn 
				      (reconnect ,store)
				      store))))
     (unwind-protect (progn 
                     ,@body)
       (disconnect *database*))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Initialization/finalization ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod open-store ((store-type (eql :postmodern)) &rest args)
  "Postmodern connect (database user password host &key (port 5432) 
pooled-p use-ssl). " 
  (setf *default-store* (apply #'connect args)))

(defmethod close-store ((store database-connection))
  (when (eq *default-store* store)
    (setf *default-store* nil))
  (disconnect store))

;;From common lisp cook book
(defun replace-all (string part replacement &key (test #'char=))
"Returns a new string in which all the occurences of the part 
is replaced with replacement."
    (with-output-to-string (out)
      (loop with part-length = (length part)
            for old-pos = 0 then (+ pos part-length)
            for pos = (search part string
                              :start2 old-pos
                              :test test)
            do (write-string string out
                             :start old-pos
                             :end (or pos (length string)))
            when pos do (write-string replacement out)
            while pos))) 


(defun list-join-to-string (list delimiter)
(string-right-trim delimiter
  (with-output-to-string (str)
    (dolist (el list)
      (write-string (replace-all (format nil "~A" el) "-" "_") str)
      (write-string delimiter str)))))

(defmethod clean-store ((store database-connection))
  "This function does a TRUNCATE but their will be a performance 
hit if there are foreign keys."
  (with-store store
      (execute 
       (format nil "TRUNCATE ~A;" (list-join-to-string (list-tables) ", ")))))

;;;;;;;;;;;;;;;;;;;;
;;; Transactions ;;;
;;;;;;;;;;;;;;;;;;;;

(defmethod use-dynamic-transaction-p ((store database-connection))
  t)

(defmethod dynamic-transaction ((store database-connection) proc)
  (with-store store      
      (funcall proc)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Creating and deleting persistent objects ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;TODO: Create a slot-class for foreign-keys because multiple values are needed. (One day when I know haow to mop.)
(defclass dao-store-class (dao-class)
  "Expected Values are for foreign-keys is a list of  (foreign-key-id-slot-name 
  foreign-key-boject-slot-name foreign-object-id-slot-name include-in-delete-p)."
((foreign-keys :initarg :foreign-keys :initform nil :reader foreign-keys)))


(defmethod class-id-slot-name ((class dao-class))
  (get-id-name))

(defun get-id-name ()
  (if *oid-p*
      'oid
      'id))

(defun get-dao-foreign (store instance class-name)
  "Retrieves foreign key instances of a dao-store-class instance."
  (when (eql (type-of (class-of instance)) 'dao-store-class)  
    (let ((f-keys (slot-value (find-class class-name) 'foreign-keys)))      
      (dolist (key-combo f-keys)
        (let ((foreign-instance (find-persistent-object-by-id store (second key-combo) (slot-value instance (first key-combo)))))                
          (when foreign-instance 
                    
            (setf (slot-value instance (second key-combo)) foreign-instance)
            ; (break (format nil "?? -->" foreign-instance)) 
            )))))
  instance)


;;TODO: Make this recurse as deep as needed.
(defun update-dao-store (store object)
  "A conveniance function if the weblocks user should ever want to update 
dao-store-class objects directly. This function updates foreign key slot 
instances as well and sets the appropriate foreign key id's."
  (when (eql (type-of (class-of object)) 'dao-store-class)
    (with-store store
        (let ((f-keys (slot-value (class-of object) 'foreign-keys)))          
          (dolist (key-combo f-keys)       
            (let ((foreign-instance 
                   (if (slot-boundp object (second key-combo))
                       (slot-value object (second key-combo)))))
              (when foreign-instance
                (save-dao foreign-instance)
                (setf (slot-value object (car key-combo)) (slot-value foreign-instance (third key-combo)))))))
      (save-dao object)))
  object)

;;TODO: Make this recurse as deep as needed.
(defun delete-dao-store (store object)
  "A conveniance function if the weblocks user should ever want to delete 
dao-store-class objects directly. This function deletes foreign key slot 
instances as well if marked for deleteion."
  (when (eql (type-of (class-of object)) 'dao-store-class)
    (with-store store
        (let ((f-keys (slot-value (class-of object) 'foreign-keys)))
          (dolist (key-combo f-keys)       
            (let ((foreign-instance 
                   (if (slot-boundp object (second key-combo))
                       (slot-value object (second key-combo)))))
              (when (and foreign-instance (fourth key-combo))
                
                (when (slot-boundp object (third key-combo))
                  (delete-dao foreign-instance)
                  (setf (slot-value object (car key-combo)) nil))))))))
      (delete-dao object)      
  object)

(defmethod persist-object ((store database-connection) object &key)
  "Tries to insert the given DAO using insert-dao. If this raises a 
unique key violation error, it tries to update it using update-dao 
instead. Be aware that there is a possible race condition here ― 
if some other process deletes the row at just the right moment, 
the update fails as well. "
  (with-store store
      (save-dao object)))

(defmethod delete-persistent-object ((store database-connection) object)
  "This method does not use cascading delete because weblocks does not expect this behavior.
Specifically the grid passes the mixin classes to this function one by one effectivly doing
a cascading delete on its own. If you need a cascading delete then use delete-dao-store. " 
  (if object
      (if (slot-boundp object (get-id-name)) ;;Need to look into the get-id thing in wblocks it works but in unit tests there is package issues???            
          (with-store store
              (delete-dao object))))
  object)


(defmethod delete-persistent-object-by-id ((store database-connection) class-name 
                                           object-id)
   "This method does not use cascading delete because weblocks does not expect this behavior.
  If you need a cascading delete then use delete-dao-store. " 
  (with-store store
    (query (:delete :from class-name :where (:= (get-id-name) object-id)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Querying persistent objects ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod find-persistent-object-by-id ((store database-connection) class-name 
                                         object-id)
  "This function retrieves foreighn key slot instances as well if they exist."  
  (with-store store
      (get-dao-foreign store
                       (car (query-dao class-name 
                                       (sql-compile (if *oid-p*
                                                        `(:limit (:select oid,'* :from ,class-name 
                                                                   :where (:= ,(get-id-name) ,object-id)) 
                                                           1)
                                                        `(:limit (:select '* :from ,class-name 
                                                                   :where (:= ,(get-id-name) ,object-id)) 
                                                           1)))))
                       class-name)))

;;Nipped from weblocks-clsql
(defun range-to-offset (range)
  "Converts the 'range' argument to SQL OFFSET."
  (when range
    (car range)))

;;Nipped from weblocks-clsql
(defun range-to-limit (range)
  "Converts the 'range' argument to SQL LIMIT."
  (when range
    (- (cdr range) (car range))))


;;http://blog.viridian-project.de/2008/02/17/string-for-the-rest-of-us/
(defun string+ (&rest parts)
  (with-output-to-string (out)
    (dolist (part parts)
      (write part :stream out :escape nil))))




;;:TODO: I think there should be a cleaner way to do this with s-sql, but I could not 
;;get the syntax right for more than one if at a time.
(defmethod find-persistent-objects ((store database-connection) class-name &key
                                    order-by range test)
  "Retrieves persistant objects with their foreign key slot instances."
  (let ((daos (with-store store
                  (query-dao class-name                                        
                             (string+ (if *oid-p*
                                          "select oid,* from "
                                          "select * from ") (string-downcase (replace-all (symbol-name class-name) "-" "_")) " "
                                      
                                      (if test
                                          (string+ " where " (format nil "~A ~%" test))
                                          "")
                                      
                                      (if order-by
                                          (string+ " order by " (format nil "~A  ~A ~%" 
                                                                        (string-downcase  (replace-all (symbol-name (car order-by)) "-" "_")) 
                                                                        (if (cdr order-by)
                                                                            (cdr order-by)
                                                                            "" )))
                                          "")                               
                                      
                                      (if range
                                          (string+ " limit " (range-to-limit range) " offset " (range-to-offset range))
                                          ""))
                             ))))
    (dolist (dao daos)
      (get-dao-foreign store dao class-name))
    daos))
              
(defmethod count-persistent-objects ((store database-connection) class-name &key
                                     test &allow-other-keys)
  (with-store store
      (query                      
       (string+ "select count(*) from " (string-downcase (replace-all (symbol-name class-name) "-" "_"))  " "
                (if test
                    (string+ " where " (format nil "~A ~%" test))
                    ""))
       :single)
    ))
  

